#include <cstdio>
#include "glinc.h"
#include <stdlib.h>
#include "geo.h"
#include "projectiles.h"
#include "utils.h"
#include "feeble.h"
#include "bleep.h"
#include <cmath>

using namespace geogame;

using namespace std;
const unsigned int MAXDESC=1024;
const int MAXGEOS=400;
int Geo::GeoCount = 0;

static const int GEO_MAX_ENERGY=200;
static const int GEO_FIRE_ENERGY=50;
static const int GEO_FIRE_COST=5;
static const int GEO_TICK_RATE=50;
static const int GEO_TICK_RATE_VAR=50;


Component::Component(const dMatrix &space) : 
m_Space(space) 
{
	m_Pos=space.transform(m_Pos);
}

void TriangleComponent::Render()
{
	Primitive::Get()->Triangle();	
}

void SquareComponent::Render()
{ 
	Primitive::Get()->Square();
}

void PentagonComponent::Render()
{
	Primitive::Get()->Pentagon();
}

dMatrix GetMatrix()
{
	dMatrix m;
	glGetFloatv(GL_MODELVIEW_MATRIX,(float*)m.m);
	return m;
}

///////////////////////////////////////////////////////////////////////////////////

bool Geo::m_InitAudio=false;
int Geo::m_ExplosionSound=0;
int Geo::m_DeathSound=0;
int Geo::m_ShotSound=0;

Geo::Geo(int generation, const dMatrix &space) :
m_Dirty(true),
m_Space(space),
m_Age(0),
m_Ticks(0),
m_Rate(GEO_TICK_RATE+rand()%GEO_TICK_RATE_VAR),
m_Energy(10),
m_Lifetime(10),
m_NumTriangles(0),
m_NumSquares(0),
m_JustRebuilt(false),
m_Mass(0)
{
	m_Flags|=ENTITYFLAG_ENEMY;
	m_Position=space.transform(m_Position);	
	
	if (m_Position.x<-200 || m_Position.x>200 || m_Position.y<-200 || m_Position.y>200)
	{
		m_RemoveMe=true;
	}

	GeoCount++;
	m_Genome.Generation=generation;
	m_Genome.Score=0;

	if (!m_InitAudio) InitAudio();
}

Geo::~Geo()
{
	for (list<Component*>::iterator i=m_Components.begin(); i!=m_Components.end(); i++)
	{
		delete *i;
	} 
	
	GeoCount--;
}

void Geo::Update(const list<Entity*> &world, const dVector &playerpos)
{
	Entity::Update(world,playerpos);
	
	m_JustRebuilt=false;
	if (m_Dirty)
	{
		for(list<Component*>::iterator i=m_Components.begin(); i!=m_Components.end(); i++)
		{
			delete *i;
		}
		m_Components.clear();
		
		m_BoundingBox.clear();
		
		glPushMatrix();
		glLoadIdentity();
		glMultMatrixf((float*)m_Space.m);
		RenderPart(m_Genome.Description, 0);
		glPopMatrix();
	
		m_BoundingBox.expandby(1.5);
		
		m_Dirty=false;
		m_JustRebuilt=true;
	}

	// kill any components we've grown into
	if (m_JustRebuilt)
	{
		int count=0;
		// should do a bb to bb check first
		for (list<Entity*>::const_iterator i=world.begin(); i!=world.end(); i++)
		{
			
			if ((*i)->GetFlags()==ENTITYFLAG_ENEMY && *i!=this) 
			{
				for(list<Component*>::iterator c=m_Components.begin(); c!=m_Components.end(); c++)
				{
					(*i)->CheckCollide((*c)->GetPosition(),0.6,true);
				}
			}
			count++;
		}
		
		// update the mass (uses max)
		if (count>m_Mass) m_Mass=count;
		
		m_NumTriangles=0;
		m_NumSquares=0;
		m_FirePos.clear();
		m_BirthPos.clear();
		
		for (list<Component*>::iterator i=m_Components.begin(); i!=m_Components.end(); i++)
		{
			if ((*i)->NumSides()==3) 
			{
				m_NumTriangles++;
				m_FirePos.push_back((*i)->GetPosition());	
			}	
			if ((*i)->NumSides()==4) m_NumSquares++;
			if ((*i)->NumSides()==5) m_BirthPos.push_back((*i)->GetSpace());
		}
	
	}	
	
	if (m_Ticks%m_Rate==0) 
	{
		m_Energy += m_NumSquares;
		if (m_Energy<GEO_MAX_ENERGY) Process();
		m_Age++;
		
		// if player in range
		if (m_Energy>GEO_FIRE_ENERGY && playerpos.distsq(m_Position)<30*30)
		{
			for(vector<dVector>::iterator i=m_FirePos.begin(); i!=m_FirePos.end(); i++)
			{
				if (m_Energy>GEO_FIRE_ENERGY) 
				{
					Bleep::Get()->Play(m_ShotSound, 220, 1);
					dVector jitter(RandRange(-5,5),RandRange(-5,5),RandRange(-5,5));
					Rocket *newrocket = new Rocket(*i,playerpos+jitter,RandRange(0.2,1));
					newrocket->SetExplodeSound(m_ExplosionSound);
					m_NewEntities.push_back(newrocket);
					m_Energy-=GEO_FIRE_COST;
					//m_Genome.Score++;
				}
			}
		}
		
		if (m_Energy==0) m_RemoveMe=true;

		if(m_Energy>GEO_MAX_ENERGY && GeoCount<MAXGEOS)
		{
			Bleep::Get()->Play(m_DeathSound, 220, 1);

			// give birth 
			for(vector<dMatrix>::iterator i=m_BirthPos.begin(); i!=m_BirthPos.end(); i++)
			{
				Geo *seed = new Geo(++m_Genome.Generation, *i);
				seed->m_Genome.Axiom=m_Genome.Axiom;
				seed->m_Genome.Rules=m_Genome.Rules;
				seed->Mutate(0.01);
				seed->m_Genome.Description=m_Genome.Axiom;
				m_NewEntities.push_back(seed);
				// encourage size
				m_Genome.Score+=(int)((m_BoundingBox.max.x-m_BoundingBox.min.x)*(m_BoundingBox.max.y-m_BoundingBox.min.y));
			}
			m_RemoveMe=true;
		}

		if (m_Components.empty()) m_RemoveMe=true;
	}
	
	m_Ticks++;
}

void Geo::RenderPart(const string &part, int depth)
{
	if (depth>=20) return;
	
	//cerr<<"RenderPart: "<<part<<endl;
	// parse string
	int numsides=0;
	unsigned int pos=0;
	Component *newcomp;
	dMatrix m=GetMatrix();
	
	bool done=false;
	while (!done)
	{
		switch (part[pos])
		{
			case 't': newcomp = new TriangleComponent(m); done=true; break;
			case 's': newcomp = new SquareComponent(m); done=true; break;
			case 'p': newcomp = new PentagonComponent(m); done=true; break;
			break;
			default: break; // invalid or malformed, or containing an lsystem remmnant
		}
		pos++; 
		if (pos>=part.size()) return; // no usable sections found
	}
	
	dVector zero(0,0,0);
	zero = m.transform(zero);
	m_BoundingBox.expand(zero);
	numsides=newcomp->NumSides(); 
	m_Components.push_back(newcomp);
	
	pos=1;	
	int side=1;	
	while(pos<part.size() && side<numsides)
	{
		string branch=ExtractBranch(part,pos);
		//cerr<<"ExtractBranch: "<<part<<" "<<pos<<": "<<branch<<endl;
		if (branch!="")
		{
			glPushMatrix();
			glRotatef(180+(side*(360.0/(float)numsides)),0,0,1);
			glTranslatef(0,0.6,0);
			RenderPart(branch,++depth);
			glPopMatrix();
		}
		side++;
	}
}

void Geo::Render()
{
	for(list<Component*>::iterator i=m_Components.begin(); i!=m_Components.end(); i++)
	{
		glPushMatrix();
		glMultMatrixf((float*)(*i)->GetSpace().m);
		(*i)->Render();
		glPopMatrix();
	}
	
	/*glPushMatrix();
	glMultMatrixf((float*)m_Space.m);
	glColor4f(1,1,1,1); 
	//char info[256];
	//sprintf(info,"age:%d gen: %d energy:%d life:%d",
	//	m_Age,m_Generation,m_Energy,m_Lifetime);
	//Text(info);
	glTranslatef(-0.4,-0.2,0);
	//Text(m_Description);	
	glPopMatrix();*/
}

string Geo::ExtractBranch(const string &ls, unsigned int &pos) // extracts bit between []
{
    int dcount=-1;
    string::const_iterator i=ls.begin();
	for (unsigned int n=0; n<pos; n++) i++;
	// search for a branch start
	while (*i!='[' && i!=ls.end()) i++;
	string::const_iterator start=i;
	start++;
	pos++;
	
    while(i!=ls.end())
    {
        if (dcount==0 && (*i==']')) 
        {
            return string(start,i);
        }
        if (*i=='[') dcount++;
        if (*i==']') dcount--;
        i++;
		pos++;
    }
	
	//cerr<<"invalid"<<endl;
    // the structure is technically invalid, but we'll just return nothing 
    // to make it less brittle (makes mutation much easier) - basically just
    // ignores the ['s with no corresponding ]
    return string("");
}

void Geo::Process()
{
	m_Dirty=true;
    string temp,replacment;
    string::iterator i=m_Genome.Description.begin();

    while(i!=m_Genome.Description.end())
    {
        replacment=*i;
        for (vector<pair<char,string> >::iterator r=m_Genome.Rules.begin(); r!=m_Genome.Rules.end(); r++) 
        {
            if (*i==r->first) 
            {     
                replacment=r->second;
            }
        }
        temp+=replacment;
		if (temp.size()>MAXDESC) break;
        i++;
    }
    m_Genome.Description=temp;
}

string Geo::GetFragment() 
{
	switch(rand()%29)
	{
		case 0: return "[";
		case 1: return "]";
		case 2: return "t";
		case 3: return "s";
		case 4: return "p";
		case 5: return "0";
		case 6: return "1";
		case 7: return "2";
		case 8: return "3";
		case 9: return "[";
		case 10: return "]";
		case 11: return "[";
		case 12: return "]";
		case 13: return "[";
		case 14: return "]";
		case 15: return "[";
		case 16: return "]";
		case 17: return "t";
		case 18: return "s";
		case 19: return "p";
		case 20: return "t";
		case 21: return "s";
		case 22: return "p";
	}
	return "";
}	

void Geo::Randomise()
{
	m_Dirty=true;

	int desclength=10;
	m_Genome.Axiom="";
	for(int n=0; n<desclength; n++)
	{
		m_Genome.Axiom+=GetFragment();
	}
	m_Genome.Description=m_Genome.Axiom;
	
	m_Genome.Rules.clear();
	int numrules=4;
	for(int r=0; r<numrules; r++)
	{
		int rulelength=10;
		string rule="";
		for(int n=0; n<desclength; n++) rule+=GetFragment();
		char temp[3];
		sprintf(temp,"%d",r);
		m_Genome.Rules.push_back(pair<char,string>(temp[0],rule));
	}
}

string Geo::MutateHelper(string s, float rate)
{
	string copy;
	for (unsigned int i=0; i<s.size(); i++)
	{
		if (RandFloat()<rate) copy+=GetFragment(); // sometimes insert something new
		if (RandFloat()>rate) copy+=s[i]; // sometimes forget to copy old part
	}
	return copy;
}

void Geo::Mutate(float rate)
{
	m_Genome.Axiom=MutateHelper(m_Genome.Axiom,rate);
	m_Genome.Description=MutateHelper(m_Genome.Description,rate);
	for (vector<pair<char,string> >::iterator i=m_Genome.Rules.begin(); i!=m_Genome.Rules.end(); i++)
	{
		i->second=MutateHelper(i->second,rate);
	}
	Process();
}

bool Geo::CheckCollide(const dVector &pos, float radius, bool damage)
{
	if (!m_BoundingBox.inside(pos,radius)) return false;
	
	for(list<Component*>::iterator i=m_Components.begin(); i!=m_Components.end(); i++)
	{
		if ((*i)->GetPosition().distsq(pos)<(1+radius)*(1+radius)) 
		{
			if (damage) 
			{
				delete *i;
				m_Components.erase(i);
			}
			return true;
		}
	}
	
	return false;
}

void Geo::InitAudio()
{
	m_InitAudio=true;
	
	Feeble::TimeLength=5;
	Feeble::Samplerate=22050;
	
	Feeble::TimeLength=2;
	Feeble::Value volvalue(0.1);
	Feeble::Mult volmult;

	// geo explosion sound
	{
	Feeble::TimeLength=0.5;
	Feeble::Envelope env(Feeble::Envelope::HILL);
	Feeble::Mult envmult;
	Feeble::Value envmultvalue(50);
	Feeble::Generator osc(Feeble::Generator::NOISE);	
	
	envmult.Patch("a",&envmultvalue);
	envmult.Patch("b",&env);
	osc.Patch("freq",&envmult);
	volmult.Patch("a",&osc);
	volmult.Patch("b",&volvalue);
	volmult.Execute();
		
	m_ExplosionSound = Bleep::Get()->Register(volmult.GetBuffer(),volmult.GetSize());
	}
	
	// die sound
	{
	Feeble::TimeLength=0.1;
	Feeble::Envelope env(Feeble::Envelope::RAMPUP);
	Feeble::Mult envmult;
	Feeble::Value envmultvalue(1000);
	Feeble::Generator osc(Feeble::Generator::SINE);	
	
	envmult.Patch("a",&envmultvalue);
	envmult.Patch("b",&env);
	osc.Patch("freq",&envmult);
	volmult.Patch("a",&osc);
	volmult.Patch("b",&volvalue);
	volmult.Execute();
		
	m_DeathSound = Bleep::Get()->Register(volmult.GetBuffer(),volmult.GetSize());
	}
	
	// shot sound
	{
	Feeble::TimeLength=0.5;
	Feeble::Envelope env(Feeble::Envelope::RAMPUP);
	Feeble::Mult envmult;
	Feeble::Value envmultvalue(2000);
	Feeble::Generator osc(Feeble::Generator::SINE);	
	
	envmult.Patch("a",&envmultvalue);
	envmult.Patch("b",&env);
	osc.Patch("freq",&envmult);
	volmult.Patch("a",&osc);
	volmult.Patch("b",&volvalue);
	volmult.Execute();
		
	m_ShotSound = Bleep::Get()->Register(volmult.GetBuffer(),volmult.GetSize());
	}

}





