#include "glinc.h"
#include "projectiles.h"
#include "geo.h"
#include "bleep.h"
#include "utils.h"

using namespace geogame;

Projectile::Projectile(const dVector &pos, const dVector &vec, const dVector &inr) :
m_Pos(pos),
m_Vec(vec),
m_Inertia(inr),
m_Age(0)
{
	m_Flags|=ENTITYFLAG_PROJECTILE;
} 

Projectile::~Projectile()
{
}

void Projectile::Update(const list<Entity*> &world, const dVector &playerpos)
{
	Entity::Update(world,playerpos);
	m_Pos+=m_Vec+m_Inertia;
	
	m_Age++;
	if (m_Age>400) m_RemoveMe=true;
}

///////////////////////////////////////////////////////////////////////

Bullet::Bullet(const dVector &pos, const dVector &vec, const dVector &inr) : 
Projectile(pos,vec,inr) 
{
}

void Bullet::Render()
{
	glBegin(GL_LINES);
	glColor3f(1,0,0);
	glVertex3f(m_Pos.x,m_Pos.y,0);
	dVector temp=m_Vec;
	temp.normalise();
	temp*=0.3;
	glVertex3f(m_Pos.x+temp.x,m_Pos.y+temp.y,0);
	glEnd();
}

void Bullet::Update(const list<Entity*> &world, const dVector &playerpos)
{
	Projectile::Update(world,playerpos);
	
	for (list<Entity*>::const_iterator i=world.begin(); i!=world.end(); i++)
	{
		if (*i!=this && (*i)->CheckCollide(m_Pos,0,true)) 
		{
			m_RemoveMe=true;
		}
	}
	if (m_Age>100) m_RemoveMe=true;
}

///////////////////////////////////////////////////////////////////////

Missile::Missile(const dVector &pos, const dVector &vec, const dVector &inr) : 
Projectile(pos,vec,inr), 
m_Exploding(false),
m_Fade(1)
{
}

void Missile::Render()
{
	if (!m_Exploding)
	{
		glBegin(GL_LINES);
		glColor3f(0,1,0);
		glLineWidth(3);
		glVertex3f(m_Pos.x,m_Pos.y,0);
		dVector temp=m_Vec;
		temp.normalise();
		temp*=0.1;
		glVertex3f(m_Pos.x+temp.x,m_Pos.y+temp.y,0);
		glEnd();
	}
	else
	{
		glColor4f(1,0.5+m_Fade/2.0f,m_Fade,m_Fade);
		glPushMatrix();
		glTranslatef(m_Pos.x,m_Pos.y,0);
		glScalef(2.0f+(1-m_Fade),2.0f+(1-m_Fade),0);
		Primitive::Get()->Explosion();
		glPopMatrix();
		m_Fade-=0.01;
		if (m_Fade<=0) m_RemoveMe=true;
	}
}

void Missile::Update(const list<Entity*> &world, const dVector &playerpos)
{
	Projectile::Update(world,playerpos);
	
	for (list<Entity*>::const_iterator i=world.begin(); i!=world.end(); i++)
	{
		if (*i!=this) 
		{
			if (!m_Exploding && (*i)->CheckCollide(m_Pos,0,true)) 
			{
				Bleep::Get()->Play(m_ExplodeSound, 220+RandRange(-10,10), 1);
				m_Exploding=true;
				m_Vec=dVector(0,0,0);
				m_Inertia=dVector(0,0,0);
				m_Flags=ENTITYFLAG_EXPLOSION;
			}
		
			if (m_Exploding) (*i)->CheckCollide(m_Pos,3,true);
		}
	}
	if (m_Age>100) m_RemoveMe=true;
}

///////////////////////////////////////////////////////////////////////

Rocket::Rocket(const dVector &pos, const dVector &aimpos, float speed) : 
Projectile(pos,dVector(0,0,0),dVector(0,0,0)), 
m_Exploding(false),
m_Fade(1),
m_AimPos(aimpos),
m_Speed(speed)
{
}

void Rocket::Render()
{
	if (!m_Exploding)
	{
		glPushMatrix();
		glTranslatef(m_Pos.x,m_Pos.y,0);
		glRotatef(m_Age*5,0,0,1);
		glColor4f(1,0,0,0.5);
		float radius=0.55;
		glBegin(GL_TRIANGLES);
		glVertex3f(sin(0.0*DEG_CONV)*0.3,cos(0.0*DEG_CONV)*0.3,0);
		glVertex3f(sin(120.0*DEG_CONV)*0.3,cos(120.0*DEG_CONV)*0.3,0);
		glVertex3f(sin(240.0*DEG_CONV)*0.3,cos(240.0*DEG_CONV)*0.3,0);
		glEnd();
		glPopMatrix();
	}
	else
	{
		glColor4f(m_Fade,0.5+m_Fade/2.0f,1,m_Fade);
		glPushMatrix();
		glTranslatef(m_Pos.x,m_Pos.y,0);
		glScalef(1+(1-m_Fade),1+(1-m_Fade),0);
		Primitive::Get()->Explosion();
		glPopMatrix();
		m_Fade-=0.01;
		if (m_Fade<=0) m_RemoveMe=true;
	}
}

void Rocket::Update(const list<Entity*> &world, const dVector &playerpos)
{
	Projectile::Update(world,playerpos);
	
	// only collide with player
	if (!m_Exploding && m_Pos.distsq(playerpos)<0.25) 
	{
		Bleep::Get()->Play(m_ExplodeSound, 220+RandRange(-10,10), 1);
		m_Exploding=true;
		m_Vec=dVector(0,0,0);
		m_Inertia=dVector(0,0,0);
		m_Flags=ENTITYFLAG_EXPLOSION;
	}
	
	if (!m_Exploding)
	{
		for(list<Entity*>::const_iterator i=world.begin(); i!=world.end(); i++)
		{
			if ((*i)->GetFlags()&ENTITYFLAG_CAUGHTKEY &&
				m_Pos.distsq((*i)->GetPosition())<1)
			{
				Bleep::Get()->Play(m_ExplodeSound, 440+RandRange(-10,10), 1);
				m_Exploding=true;
				m_Vec=dVector(0,0,0);
				m_Inertia=dVector(0,0,0);
				m_Flags=ENTITYFLAG_EXPLOSION;					
			}
		}
	}
	
	if (!m_Exploding)
	{	
		dVector aim=playerpos-m_Pos;
		aim.normalise();
		//aim*=0.05;
		aim*=0.005;
		m_Vec+=aim;
		//m_Vec.normalise();
		//m_Vec*=m_Speed;
	}
	
	if (m_Age>100 && !m_Exploding) m_RemoveMe=true;
}

bool Rocket::CheckCollide(const dVector &pos, float radius, bool damage)
{
	return false;
}

///////////////////////////////////////////////////////////////////////

Mutator::Mutator(const dVector &pos, const dVector &vec, const dVector &inr) : 
Projectile(pos,vec,inr), 
m_Exploding(false),
m_Size(1)
{
}

void Mutator::Render()
{
	if (!m_Exploding)
	{
		glBegin(GL_LINES);
		glColor3f(0,1,0);
		glVertex3f(m_Pos.x,m_Pos.y,0);
		dVector temp=m_Vec;
		temp.normalise();
		temp*=0.1;
		glVertex3f(m_Pos.x+temp.x,m_Pos.y+temp.y,0);
		glEnd();
	}
	else
	{
		float col=1-m_Size/5.0f;
		glColor4f(RandFloat(),RandFloat(),RandFloat(),0.5);
		glPushMatrix();
		glTranslatef(m_Pos.x,m_Pos.y,0);
		glScalef(m_Size,m_Size,0);
		Primitive::Get()->Explosion();
		glPopMatrix();
		m_Size+=1;
		if (m_Size>=10) m_RemoveMe=true;
	}
}

void Mutator::Update(const list<Entity*> &world, const dVector &playerpos)
{
	Projectile::Update(world,playerpos);
	
	for (list<Entity*>::const_iterator i=world.begin(); i!=world.end(); i++)
	{
		if (*i!=this) 
		{
			if (!m_Exploding && (*i)->CheckCollide(m_Pos,0,true)) 
			{
				Bleep::Get()->Play(m_ExplodeSound, 220+RandRange(-10,10), 1);
				m_Exploding=true;
				m_Vec=dVector(0,0,0);
				m_Inertia=dVector(0,0,0);
				m_Flags=ENTITYFLAG_EXPLOSION;
			}
		
			if (m_Exploding && 
				(*i)->GetFlags()&ENTITYFLAG_ENEMY && 
				(*i)->CheckCollide(m_Pos,m_Size,false)) 
			{	
				Geo* geo=(Geo*)*i;
				geo->Mutate(0.1);
			}
		}
	}
}
